#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Naca 0012 model
# Adrien Crovato

import fpm
import tbox
import tbox.gmsh as gmsh
from fwk.coloring import ccolors

def main(p, tail=False, field=False):
    # timer
    tms = fpm.Timers()
    tms['total'].start()
    # mesh
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    msh = gmsh.MeshLoader('n0012.geo', __file__).execute(**p['pars'])
    # problem
    pbl = fpm.Problem(msh, p['aoa'], p['aos'], p['mach'], p['spn'], 1, 0, 0, 0, bool(p['sym']))
    wing = fpm.Body(msh, 'wing', ['wTe'], 10)
    pbl.add(wing)
    if tail:
        tail = fpm.Body(msh, 'tail', ['tTe'], 10)
        pbl.add(tail)
    if field:
        field = fpm.Field(msh, 'field')
        pbl.set(field)
    # AIC builder
    aic = fpm.Builder(pbl)
    aic.run()
    # solver
    slv = fpm.Solver(aic)
    slv.run()
    slv.save(tbox.GmshExport(msh))
    print('\n', slv)
    # end timer
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print(tms)
    # return
    return slv

if __name__ == '__main__':
    main()

