/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fProblem.h"
#include "fBuilder.h"
#include "fField.h"
#include "fBody.h"
#include "fWake.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include "wCache.h"
using namespace tbox;
using namespace fpm;

Problem::Problem(std::shared_ptr<MshData> _msh, double aoa, double aos, double mch,
                 double sref, double cref, double xref, double yref, double zref, bool symy) : msh(_msh), alpha(aoa), beta(aos), mach(mch), sR(sref), cR(cref), symY(symy)
{
    xR(0) = xref;
    xR(1) = yref;
    xR(2) = zref;
}

/**
 * @brief Set the fluid field
 */
void Problem::set(std::shared_ptr<Field> f)
{
    field = f;
}

/**
 * @brief Add a non-lifting body
 */
void Problem::add(std::shared_ptr<Body> b)
{
    bodies.push_back(b);
}

/**
 * @brief Compute velocity at element
 */
Eigen::Vector3d Problem::U(tbox::Element &e, std::vector<double> const &mu, double const &tau)
{
    Eigen::Vector3d U(0, 0, 0);
    size_t nVertices = e.nodes.size();             // Number of nodes per panel
    size_t nGP = e.getVCache().getVGauss().getN(); // Number of Gauss points on the element
    Eigen::MatrixXd coords(3, nVertices);
    for (int i = 0; i < 3; ++i)
        for (int j = 0; j < nVertices; ++j)
            coords(i, j) = e.nodes[(nVertices - 1) - j]->pos[i]; // Coordinates of the panel

    // Import of the shape functions at the Gauss points
    Eigen::MatrixXd ff(nVertices, nGP);
    for (int i = 0; i < nVertices; ++i)
        for (int j = 0; j < nGP; ++j)
            ff(i, j) = e.getVCache().getSf(j)[i];

    // Import of the derivatives of the shape functions at the Gauss points
    Eigen::MatrixXd dff(2 * nGP, nVertices);
    for (int i = 0; i < nGP; ++i)
        for (int j = 0; j < 2; ++j)
            dff.row(j + (2 * i)) = e.getVCache().getDsf(i).row(j);

    // Computation of the Jacobian matrix - xi-derivative (first row of the Jacobian)
    Eigen::MatrixXd dXi(3, nGP);
    for (int i = 0; i < nGP; ++i)
    {
        dXi.col(i) << 0, 0, 0;

        for (int j = 0; j < nVertices; ++j)
        {
            dXi(0, i) += dff((2 * i), j) * coords(0, j);
            dXi(1, i) += dff((2 * i), j) * coords(1, j);
            dXi(2, i) += dff((2 * i), j) * coords(2, j);
        }
    }

    // Computation of the Jacobian matrix - eta-derivative (second row of the Jacobian)
    Eigen::MatrixXd dEta(3, nGP);
    for (int i = 0; i < nGP; ++i)
    {
        dEta.col(i) << 0, 0, 0;

        for (int j = 0; j < nVertices; ++j)
        {
            dEta(0, i) += dff(1 + (2 * i), j) * coords(0, j);
            dEta(1, i) += dff(1 + (2 * i), j) * coords(1, j);
            dEta(2, i) += dff(1 + (2 * i), j) * coords(2, j);
        }
    }

    // Computation of the Jacobian matrix - cross-product (third row of the Jacobian)
    Eigen::MatrixXd dZeta(3, nGP);

    for (int i = 0; i < nGP; ++i)
    {
        Eigen::Vector3d dxi = dXi.col(i);
        Eigen::Vector3d deta = dEta.col(i);
        Eigen::Vector3d dzeta;
        dzeta = (dxi.cross(deta)).normalized();
        dZeta.col(i) = dzeta;
    }

    // Assembly of the Jacobian and partial computation of the velocity
    // Done for each Gauss point -> For n Gauss points, there are n values of the velocity -> Average
    Eigen::MatrixXd v(3, nGP);
    for (int i = 0; i < nGP; ++i)
    {
        // Creation of the Jacobian
        Eigen::Matrix3d J0;
        J0.row(0) = dXi.col(i).transpose();
        J0.row(1) = dEta.col(i).transpose();
        J0.row(2) = dZeta.col(i).transpose();

        // Inversion of the Jacobian
        Eigen::Matrix3d Jinv = J0.inverse();

        // Removal of the third column
        Eigen::MatrixXd J(3, 2);
        for (int j = 0; j < 2; ++j)
        {
            J.col(j) = Jinv.col(j);
        }

        // Import of the doublets at the nodes of the panel
        Eigen::VectorXd mu_element(nVertices);
        for (int j = 0; j < nVertices; ++j)
        {
            mu_element(j) = mu[e.nodes[j]->no - 1];
        }

        // Computation of the velocity
        Eigen::MatrixXd dff0(2, 4);
        dff0.row(0) = dff.row(2 * i);
        dff0.row(1) = dff.row(1 + (2 * i));
        v.col(i) = J * dff0 * mu_element;
    }

    // Average of the velocity
    for (int i = 0; i < nGP; ++i)
    {
        U += v.col(i);
    }
    U = U / nGP;

    // Import of the unit normal vector - Target panel
    Eigen::Vector3d n = e.getNormal();

    // Contribution of the normal velocity
    U += (-tau) * n;

    // Contribution of the freestream velocity
    U += Ui();

    return U;
}

/**
 * @brief Compute density at element
 */
double Problem::Rho(Eigen::Vector3d U)
{
    // particularized: pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), 1 / (gamma - 1));
    Eigen::Vector3d u = U;
    double a = 1 + 0.2 * mach * mach * (1 - u.dot(u));
    return sqrt(a * a * a * a * a);
}

/**
 * @brief Compute Mach number at element
 */
double Problem::Mach(Eigen::Vector3d U)
{
    if (mach == 0)
        return 0;
    else
    {
        // particularized: pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), 1 / (gamma - 1));
        Eigen::Vector3d u = U;
        double a2 = 1 / (mach * mach) + 0.2 * (1 - u.dot(u)); // speed of sound squared
        return u.norm() / sqrt(a2);
    }
}

/**
 * @brief Compute pressure coefficient at element
 */
double Problem::Cp(Eigen::Vector3d U)
{
    Eigen::Vector3d u = U;
    if (mach == 0)
        return 1 - u.dot(u);
    else
    {
        //particularized: 2 / (gamma * mInf * mInf) * (pow(1 + 0.5 * (gamma - 1) * mInf * mInf * (1 - gradU2), gamma / (gamma - 1)) - 1);
        double a = 1 + 0.2 * mach * mach * (1 - u.dot(u));
        return 2 / (1.4 * mach * mach) * (sqrt(a * a * a * a * a * a * a) - 1);
    }
}

/**
 * @brief Update the elements memory (Jacobian)
 */
void Problem::updateMem()
{
    // Update volume Jacobian
    if (field)
        for (auto e : field->tag->elems)
            e->update();
    // Update surface Jacobian
    for (auto s : bodies)
    {
        for (auto e : s->tag->elems)
        {
            e->update();
        }
        for (auto w : s->wakes)
            for (auto e : w->tag->elems)
                e->update();
    }
}

/**
 * @brief Check that Problem is not empty and that elements are supported
 */
void Problem::check() const
{
    // Sanity checks
    if (bodies.empty())
        throw std::runtime_error("No lifting bodies provided!\n");
    if (mach != 0 && !field)
    {
        std::stringstream err;
        err << "Compressible computation requested for Mach "
            << mach << ", but no field provided!\n";
        throw std::runtime_error(err.str());
    }
    if (mach == 0 && field)
        std::cout << "Incompressible computation requested, but a field was provided. This will result in unnecessary computations!\n"
                  << std::endl;

    // Three-dimension problem
    // Volume elements
    if (field)
        for (auto e : field->tag->elems)
            if (e->type() != ELTYPE::HEX8)
            {
                std::stringstream err;
                err << "FPM solver is only implemented for surface elements of type "
                    << ELTYPE::HEX8 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
    // Surface elements
    for (auto b : bodies)
        for (auto e : b->tag->elems)
            if (e->type() != ELTYPE::QUAD4)
            {
                std::stringstream err;
                err << "FPM solver is only implemented for surface elements of type "
                    << ELTYPE::QUAD4 << " (" << e->type() << " was given)!\n";
                throw std::runtime_error(err.str());
            }
}

void Problem::write(std::ostream &out) const
{
    out << "fpm::Problem definition"
        << "\n\tAngle of attack: " << alpha
        << "\n\tAngle of sideslip: " << beta
        << "\n\tReference area: " << sR
        << "\n\tReference length: " << cR
        << "\n\tReference point: [" << xR.transpose() << "]"
        << std::endl;
    out << "with";
    for (auto b : bodies)
        out << "\t" << *b;
}
